@@ -1,6 +1,7 @@
 # Orca
 #
 # Copyright 2009 Sun Microsystems Inc.
+# Copyright 2015-2016 Igalia, S.L.
 #
 # This library is free software; you can redistribute it and/or
 # modify it under the terms of the GNU Lesser General Public
@@ -22,22 +23,23 @@
 __id__        = "$Id:$"
 __version__   = "$Revision:$"
 __date__      = "$Date:$"
-__copyright__ = "Copyright (c) 2009 Sun Microsystems Inc."
+__copyright__ = "Copyright (c) 2009 Sun Microsystems Inc." \
+                "Copyright (c) 2015-2016 Igalia, S.L."
 __license__   = "LGPL"
 
+import collections
+import pyatspi
 import sys
 import time
 import traceback
-
-import pyatspi
+from gi.repository import Atspi, Atk
 
 from . import braille
 from . import debug
 from . import messages
 from . import object_properties
 from . import settings
-
-import collections
+from . import settings_manager
 
 def _formatExceptionInfo(maxTBlevel=5):
     cla, exc, trbk = sys.exc_info()
@@ -58,6 +60,8 @@ def _formatExceptionInfo(maxTBlevel=5):
 #
 METHOD_PREFIX = "_generate"
 
+_settingsManager = settings_manager.getManager()
+
 class Generator:
     """Takes accessible objects and generates a presentation for those
     objects.  See the generate method, which is the primary entry
@@ -71,6 +75,7 @@ class Generator:
 
         self._mode = mode
         self._script = script
+        self._activeProgressBars = {}
         self._methodsDict = {}
         for method in \
             [z for z in [getattr(self, y).__get__(self, self.__class__) for y in [x for x in dir(self) if x.startswith(METHOD_PREFIX)]] if isinstance(z, collections.Callable)]:
@@ -96,7 +101,7 @@ class Generator:
         # for the results of all the legal method names.
         #
         globalsDict = {}
-        for key in list(self._methodsDict.keys()):
+        for key in self._methodsDict.keys():
             globalsDict[key] = []
         self._addGlobals(globalsDict)
 
@@ -151,10 +156,13 @@ class Generator:
     def generateContents(self, contents, **args):
         return []
 
+    def generateContext(self, obj, **args):
+        return []
+
     def generate(self, obj, **args):
         """Returns an array of strings (and possibly voice and audio
-        specifications) that represent the complete presentatin for the
-        object.  The presentatin to be generated depends highly upon the
+        specifications) that represent the complete presentation for the
+        object.  The presentation to be generated depends highly upon the
         formatting strings in formatting.py.
 
         args is a dictionary that may contain any of the following:
@@ -173,6 +181,12 @@ class Generator:
         - forceTutorial: boolean (default=False) which says if we
           should force a tutorial to be spoken or not
         """
+
+        if self._script.utilities.isDead(obj):
+            msg = 'ERROR: Cannot generate presentation dead obj'
+            debug.println(debug.LEVEL_INFO, msg, True)
+            return []
+
         startTime = time.time()
         result = []
         globalsDict = {}
@@ -253,6 +267,9 @@ class Generator:
         for element in result:
             debug.println(debug.LEVEL_ALL, "           %s" % element)
 
+        if args.get('isProgressBarUpdate') and result:
+            self.setProgressBarUpdateTimeAndValue(obj)
+
         return result
 
     #####################################################################
@@ -285,6 +302,7 @@ class Generator:
         needed a _generateDescription for whereAmI. :-) See below.
         """
         result = []
+        self._script.pointOfReference['usedDescriptionForName'] = False
         name = self._script.utilities.displayedText(obj)
         if obj.getRole() == pyatspi.ROLE_COMBO_BOX:
             children = self._script.utilities.selectedChildren(obj)
@@ -349,10 +367,15 @@ class Generator:
         result = []
         label = self._generateLabel(obj, **args)
         name = self._generateName(obj, **args)
+        role = args.get('role', obj.role)
+        if not (label or name) and role == pyatspi.ROLE_TABLE_CELL:
+            descendant = self._script.utilities.realActiveDescendant(obj)
+            name = self._generateName(descendant)
+
         result.extend(label)
         if not len(label):
             result.extend(name)
-        elif len(name) and name[0].strip() != label[0].strip():
+        elif len(name) and name[0].split() != label[0].split():
             result.extend(name)
         return result
 
@@ -379,24 +402,36 @@ class Generator:
 
         role = args.get('role', obj.getRole())
 
-        # The text in the description is the same as the text in the page
-        # tab and similar to (and sometimes the same as) the prompt.
-        if role == pyatspi.ROLE_TERMINAL:
-            return []
-
         # Unity Panel Service menubar items are labels which claim focus and
         # have an accessible description of the text + underscore symbol used
         # to create the mnemonic. We'll work around that here for now.
         if role == pyatspi.ROLE_LABEL:
             return []
 
+        try:
+            name = obj.name
+            description = obj.description
+        except:
+            msg = "ERROR: Exception getting name and description for %s" % obj
+            debug.println(debug.LEVEL_INFO, msg, True)
+            name = ""
+            description = ""
+
+        if role == pyatspi.ROLE_ICON:
+            name = self._script.utilities.displayedText(obj) or ""
+
         result = []
-        if obj.description:
+        if description:
             label = self._script.utilities.displayedLabel(obj) or ""
-            name = obj.name or ""
-            desc = obj.description.lower()
+            desc = description.lower()
             if not (desc in name.lower() or desc in label.lower()):
                 result.append(obj.description)
+
+        if not result:
+            desc = self._script.utilities.displayedDescription(obj)
+            if desc:
+                result.append(desc)
+
         return result
 
     def _generateLabel(self, obj, **args):
@@ -459,6 +494,32 @@ class Generator:
             result.append(self._script.formatting.getString(**args))
         return result
 
+    def _generateInvalid(self, obj, **args):
+        error = self._script.utilities.getError(obj)
+        if not error:
+            return []
+
+        result = []
+        if not args.get('mode', None):
+            args['mode'] = self._mode
+        args['stringType'] = 'invalid'
+        indicators = self._script.formatting.getString(**args)
+
+        if error == 'spelling':
+            indicator = indicators[1]
+        elif error == 'grammar':
+            indicator = indicators[2]
+        else:
+            indicator = indicators[0]
+
+        errorMessage = self._script.utilities.getErrorMessage(obj)
+        if errorMessage:
+            result.append("%s: %s" % (indicator, errorMessage))
+        else:
+            result.append(indicator)
+
+        return result
+
     def _generateRequired(self, obj, **args):
         """Returns an array of strings for use by speech and braille that
         represent the required state of the object, but only if it is
@@ -558,6 +619,20 @@ class Generator:
 
         return []
 
+    def _generateSwitchState(self, obj, **args):
+        result = []
+        if not args.get('mode', None):
+            args['mode'] = self._mode
+        args['stringType'] = 'switch'
+        indicators = self._script.formatting.getString(**args)
+        state = obj.getState()
+        if state.contains(pyatspi.STATE_CHECKED) \
+           or state.contains(pyatspi.STATE_PRESSED):
+            result.append(indicators[1])
+        else:
+            result.append(indicators[0])
+        return result
+
     def _generateToggleState(self, obj, **args):
         """Returns an array of strings for use by speech and braille that
         represent the checked state of the object.  This is typically
@@ -603,11 +678,13 @@ class Generator:
         args['stringType'] = 'expansion'
         indicators = self._script.formatting.getString(**args)
         state = obj.getState()
-        if state.contains(pyatspi.STATE_EXPANDABLE):
-            if state.contains(pyatspi.STATE_EXPANDED):
-                result.append(indicators[1])
-            else:
-                result.append(indicators[0])
+        if state.contains(pyatspi.STATE_COLLAPSED):
+            result.append(indicators[0])
+        elif state.contains(pyatspi.STATE_EXPANDED):
+            result.append(indicators[1])
+        elif state.contains(pyatspi.STATE_EXPANDABLE):
+            result.append(indicators[0])
+
         return result
 
     def _generateMultiselectableState(self, obj, **args):
@@ -646,7 +723,7 @@ class Generator:
         if not text:
             return result
 
-        roleString =  self.getLocalizedRoleName(obj, pyatspi.ROLE_ROW_HEADER)
+        roleString =  self.getLocalizedRoleName(obj, role=pyatspi.ROLE_ROW_HEADER)
         if args.get('mode') == 'speech':
             if settings.speechVerbosityLevel == settings.VERBOSITY_LEVEL_VERBOSE \
                and not args.get('formatType') in ['basicWhereAmI', 'detailedWhereAmI']:
@@ -672,7 +749,7 @@ class Generator:
         if not text:
             return result
 
-        roleString =  self.getLocalizedRoleName(obj, pyatspi.ROLE_COLUMN_HEADER)
+        roleString =  self.getLocalizedRoleName(obj, role=pyatspi.ROLE_COLUMN_HEADER)
         if args.get('mode') == 'speech':
             if settings.speechVerbosityLevel == settings.VERBOSITY_LEVEL_VERBOSE \
                and not args.get('formatType') in ['basicWhereAmI', 'detailedWhereAmI']:
@@ -877,7 +954,7 @@ class Generator:
         if start is None or end is None:
             return []
 
-        substring = self._script.utilities.substring(obj, start, end)
+        substring = args.get('string', self._script.utilities.substring(obj, start, end))
         if substring and not self._script.EMBEDDED_OBJECT_CHARACTER in substring:
             return [substring]
 
@@ -963,7 +1040,7 @@ class Generator:
 
     def _generateApplicationName(self, obj, **args):
         """Returns an array of strings for use by speech and braille that
-        represents the name of the applicaton for the object.
+        represents the name of the application for the object.
         """
         result = []
         try:
@@ -1034,7 +1111,12 @@ class Generator:
         """
         result = []
         rad = self._script.utilities.realActiveDescendant(obj)
-        return self._generateDisplayedText(rad, **args)
+
+        if not (rad.getRole() == pyatspi.ROLE_TABLE_CELL and rad.childCount):
+            return self._generateDisplayedText(rad, **args)
+
+        content = set([self._script.utilities.displayedText(x).strip() for x in rad])
+        return [" ".join(filter(lambda x: x, content))]
 
     def _generateRealActiveDescendantRoleName(self, obj, **args ):
         """Objects, such as tables and trees, can represent individual cells
@@ -1070,12 +1152,57 @@ class Generator:
     def _generateProgressBarValue(self, obj, **args):
         return []
 
+    def _getProgressBarUpdateInterval(self):
+        return int(_settingsManager.getSetting('progressBarUpdateInterval'))
+
+    def _shouldPresentProgressBarUpdate(self, obj, **args):
+        percent = self._script.utilities.getValueAsPercent(obj)
+        lastTime, lastValue = self.getProgressBarUpdateTimeAndValue(obj, type=self)
+        if percent == lastValue:
+            return False
+
+        if percent == 100:
+            return True
+
+        interval = int(time.time() - lastTime)
+        return interval >= self._getProgressBarUpdateInterval()
+
+    def _cleanUpCachedProgressBars(self):
+        isValid = lambda x: not (self._script.utilities.isZombie(x) or self._script.utilities.isDead(x))
+        bars = list(filter(isValid, self._activeProgressBars))
+        self._activeProgressBars = {x:self._activeProgressBars.get(x) for x in bars}
+
+    def _getMostRecentProgressBarUpdate(self):
+        self._cleanUpCachedProgressBars()
+        if not self._activeProgressBars.values():
+            return None, 0.0, None
+
+        sortedValues = sorted(self._activeProgressBars.values(), key=lambda x: x[0])
+        prevTime, prevValue = sortedValues[-1]
+        return list(self._activeProgressBars.keys())[-1], prevTime, prevValue
+
+    def getProgressBarNumberAndCount(self, obj):
+        self._cleanUpCachedProgressBars()
+        if not obj in self._activeProgressBars:
+            self._activeProgressBars[obj] = 0.0, None
+
+        thisValue = self.getProgressBarUpdateTimeAndValue(obj)
+        index = list(self._activeProgressBars.values()).index(thisValue)
+        return index + 1, len(self._activeProgressBars)
+
+    def getProgressBarUpdateTimeAndValue(self, obj, **args):
+        if not obj in self._activeProgressBars:
+            self._activeProgressBars[obj] = 0.0, None
+
+        return self._activeProgressBars.get(obj)
+
+    def setProgressBarUpdateTimeAndValue(self, obj, lastTime=None, lastValue=None):
+        lastTime = lastTime or time.time()
+        lastValue = lastValue or self._script.utilities.getValueAsPercent(obj)
+        self._activeProgressBars[obj] = lastTime, lastValue
+
     def _getAlternativeRole(self, obj, **args):
         if self._script.utilities.isMath(obj):
-            if self._script.utilities.isMathFraction(obj):
-                return 'ROLE_MATH_FRACTION'
-            if self._script.utilities.isMathRoot(obj):
-                return 'ROLE_MATH_ROOT'
             if self._script.utilities.isMathSubOrSuperScript(obj):
                 return 'ROLE_MATH_SCRIPT_SUBSUPER'
             if self._script.utilities.isMathUnderOrOverScript(obj):
@@ -1090,7 +1217,203 @@ class Generator:
                 return 'ROLE_MATH_TABLE'
             if self._script.utilities.isMathTableRow(obj):
                 return 'ROLE_MATH_TABLE_ROW'
-        if self._script.utilities.isStatic(obj):
-            return 'ROLE_STATIC'
+        if self._script.utilities.isDPub(obj):
+            if self._script.utilities.isDPubFootnote(obj):
+                return 'ROLE_FOOTNOTE'
+            if self._script.utilities.isLandmark(obj):
+                return 'ROLE_DPUB_LANDMARK'
+            if obj.getRole() == pyatspi.ROLE_SECTION:
+                return 'ROLE_DPUB_SECTION'
+        if self._script.utilities.isSwitch(obj):
+            return 'ROLE_SWITCH'
+        if self._script.utilities.isAnchor(obj):
+            return pyatspi.ROLE_STATIC
+        if self._script.utilities.isBlockquote(obj):
+            return pyatspi.ROLE_BLOCK_QUOTE
+        if self._script.utilities.isLandmark(obj):
+            return pyatspi.ROLE_LANDMARK
+        if self._script.utilities.isFocusableLabel(obj):
+            return pyatspi.ROLE_LIST_ITEM
 
         return args.get('role', obj.getRole())
+
+    def getLocalizedRoleName(self, obj, **args):
+        role = args.get('role', obj.getRole())
+
+        if "Value" in pyatspi.listInterfaces(obj):
+            state = obj.getState()
+            isVertical = state.contains(pyatspi.STATE_VERTICAL)
+            isHorizontal = state.contains(pyatspi.STATE_HORIZONTAL)
+            isFocused = state.contains(pyatspi.STATE_FOCUSED) \
+                        or args.get('alreadyFocused', False)
+
+            if role == pyatspi.ROLE_SLIDER:
+                if isHorizontal:
+                    return object_properties.ROLE_SLIDER_HORIZONTAL
+                if isVertical:
+                    return object_properties.ROLE_SLIDER_VERTICAL
+            elif role == pyatspi.ROLE_SCROLL_BAR:
+                if isHorizontal:
+                    return object_properties.ROLE_SCROLL_BAR_HORIZONTAL
+                if isVertical:
+                    return object_properties.ROLE_SCROLL_BAR_VERTICAL
+            elif role == pyatspi.ROLE_SEPARATOR:
+                if isHorizontal:
+                    return object_properties.ROLE_SPLITTER_HORIZONTAL
+                if isVertical:
+                    return object_properties.ROLE_SPLITTER_VERTICAL
+            elif role == pyatspi.ROLE_SPLIT_PANE and isFocused:
+                # The splitter has the opposite orientation of the split pane.
+                if isHorizontal:
+                    return object_properties.ROLE_SPLITTER_VERTICAL
+                if isVertical:
+                    return object_properties.ROLE_SPLITTER_HORIZONTAL
+
+        if self._script.utilities.isFeed(obj):
+            return object_properties.ROLE_FEED
+
+        if self._script.utilities.isFigure(obj):
+            return object_properties.ROLE_FIGURE
+
+        if self._script.utilities.isMenuButton(obj):
+            return object_properties.ROLE_MENU_BUTTON
+
+        if self._script.utilities.isSwitch(obj):
+            return object_properties.ROLE_SWITCH
+
+        if self._script.utilities.isDPub(obj):
+            if self._script.utilities.isLandmark(obj):
+                if self._script.utilities.isDPubAcknowledgments(obj):
+                    return object_properties.ROLE_ACKNOWLEDGMENTS
+                if self._script.utilities.isDPubAfterword(obj):
+                    return object_properties.ROLE_AFTERWORD
+                if self._script.utilities.isDPubAppendix(obj):
+                    return object_properties.ROLE_APPENDIX
+                if self._script.utilities.isDPubBibliography(obj):
+                    return object_properties.ROLE_BIBLIOGRAPHY
+                if self._script.utilities.isDPubChapter(obj):
+                    return object_properties.ROLE_CHAPTER
+                if self._script.utilities.isDPubConclusion(obj):
+                    return object_properties.ROLE_CONCLUSION
+                if self._script.utilities.isDPubCredits(obj):
+                    return object_properties.ROLE_CREDITS
+                if self._script.utilities.isDPubEndnotes(obj):
+                    return object_properties.ROLE_ENDNOTES
+                if self._script.utilities.isDPubEpilogue(obj):
+                    return object_properties.ROLE_EPILOGUE
+                if self._script.utilities.isDPubErrata(obj):
+                    return object_properties.ROLE_ERRATA
+                if self._script.utilities.isDPubForeword(obj):
+                    return object_properties.ROLE_FOREWORD
+                if self._script.utilities.isDPubGlossary(obj):
+                    return object_properties.ROLE_GLOSSARY
+                if self._script.utilities.isDPubIndex(obj):
+                    return object_properties.ROLE_INDEX
+                if self._script.utilities.isDPubIntroduction(obj):
+                    return object_properties.ROLE_INTRODUCTION
+                if self._script.utilities.isDPubPagelist(obj):
+                    return object_properties.ROLE_PAGELIST
+                if self._script.utilities.isDPubPart(obj):
+                    return object_properties.ROLE_PART
+                if self._script.utilities.isDPubPreface(obj):
+                    return object_properties.ROLE_PREFACE
+                if self._script.utilities.isDPubPrologue(obj):
+                    return object_properties.ROLE_PROLOGUE
+                if self._script.utilities.isDPubToc(obj):
+                    return object_properties.ROLE_TOC
+            elif role == "ROLE_DPUB_SECTION":
+                if self._script.utilities.isDPubAbstract(obj):
+                    return object_properties.ROLE_ABSTRACT
+                if self._script.utilities.isDPubColophon(obj):
+                    return object_properties.ROLE_COLOPHON
+                if self._script.utilities.isDPubCredit(obj):
+                    return object_properties.ROLE_CREDIT
+                if self._script.utilities.isDPubDedication(obj):
+                    return object_properties.ROLE_DEDICATION
+                if self._script.utilities.isDPubEpigraph(obj):
+                    return object_properties.ROLE_EPIGRAPH
+                if self._script.utilities.isDPubExample(obj):
+                    return object_properties.ROLE_EXAMPLE
+                if self._script.utilities.isDPubPullquote(obj):
+                    return object_properties.ROLE_PULLQUOTE
+                if self._script.utilities.isDPubQna(obj):
+                    return object_properties.ROLE_QNA
+            elif role == pyatspi.ROLE_LIST_ITEM:
+                if self._script.utilities.isDPubBiblioentry(obj):
+                    return object_properties.ROLE_BIBLIOENTRY
+                if self._script.utilities.isDPubEndnote(obj):
+                    return object_properties.ROLE_ENDNOTE
+            else:
+                if self._script.utilities.isDPubCover(obj):
+                    return object_properties.ROLE_COVER
+                if self._script.utilities.isDPubFootnote(obj):
+                    return object_properties.ROLE_FOOTNOTE
+                if self._script.utilities.isDPubPagebreak(obj):
+                    return object_properties.ROLE_PAGEBREAK
+                if self._script.utilities.isDPubSubtitle(obj):
+                    return object_properties.ROLE_SUBTITLE
+
+        if self._script.utilities.isLandmark(obj):
+            if self._script.utilities.isLandmarkBanner(obj):
+                return object_properties.ROLE_LANDMARK_BANNER
+            if self._script.utilities.isLandmarkComplementary(obj):
+                return object_properties.ROLE_LANDMARK_COMPLEMENTARY
+            if self._script.utilities.isLandmarkContentInfo(obj):
+                return object_properties.ROLE_LANDMARK_CONTENTINFO
+            if self._script.utilities.isLandmarkMain(obj):
+                return object_properties.ROLE_LANDMARK_MAIN
+            if self._script.utilities.isLandmarkNavigation(obj):
+                return object_properties.ROLE_LANDMARK_NAVIGATION
+            if self._script.utilities.isLandmarkRegion(obj):
+                return object_properties.ROLE_LANDMARK_REGION
+            if self._script.utilities.isLandmarkSearch(obj):
+                return object_properties.ROLE_LANDMARK_SEARCH
+            if self._script.utilities.isLandmarkForm(obj):
+                role = pyatspi.ROLE_FORM
+
+        if not isinstance(role, (pyatspi.Role, Atspi.Role)):
+            try:
+                return obj.getLocalizedRoleName()
+            except:
+                return ''
+
+        nonlocalized = Atspi.role_get_name(role)
+        atkRole = Atk.role_for_name(nonlocalized)
+        if atkRole == Atk.Role.INVALID and role == pyatspi.ROLE_STATUS_BAR:
+            atkRole = Atk.role_for_name("statusbar")
+
+        return Atk.role_get_localized_name(atkRole)
+
+    def getStateIndicator(self, obj, **args):
+        if self._script.utilities.isSwitch(obj):
+            return self._generateSwitchState(obj, **args)
+
+        role = args.get('role', obj.getRole())
+
+        if role == pyatspi.ROLE_MENU_ITEM:
+            return self._generateMenuItemCheckedState(obj, **args)
+
+        if role in [pyatspi.ROLE_RADIO_BUTTON, pyatspi.ROLE_RADIO_MENU_ITEM]:
+            return self._generateRadioState(obj, **args)
+
+        if role in [pyatspi.ROLE_CHECK_BOX, pyatspi.ROLE_CHECK_MENU_ITEM]:
+            return self._generateCheckedState(obj, **args)
+
+        if role == pyatspi.ROLE_TOGGLE_BUTTON:
+            return self._generateToggleState(obj, **args)
+
+        if role == pyatspi.ROLE_TABLE_CELL:
+            return self._generateCellCheckedState(obj, **args)
+
+        return []
+
+    def getValue(self, obj, **args):
+        role = args.get('role', obj.getRole())
+
+        if role == pyatspi.ROLE_PROGRESS_BAR:
+            return self._generateProgressBarValue(obj, **args)
+
+        if role in [pyatspi.ROLE_SCROLL_BAR, pyatspi.ROLE_SLIDER]:
+            return self._generatePercentage(obj, **args)
+
+        return []
