@@ -29,6 +29,7 @@ __copyright__ = "Copyright (c) 2004-2009 Sun Microsystems Inc." \
                 "Copyright (c) 2012 Igalia, S.L."
 __license__   = "LGPL"
 
+import faulthandler
 import gi
 import importlib
 import os
@@ -69,6 +70,7 @@ from . import event_manager
 from . import keybindings
 from . import logger
 from . import messages
+from . import mouse_review
 from . import notification_messages
 from . import orca_state
 from . import orca_platform
@@ -78,19 +80,12 @@ from . import settings_manager
 from . import speech
 from . import sound
 from .input_event import BrailleEvent
-from .input_event import KeyboardEvent
 
 _eventManager = event_manager.getManager()
 _scriptManager = script_manager.getManager()
 _settingsManager = settings_manager.getManager()
 _logger = logger.getLogger()
 
-try:
-    # If we don't have an active desktop, we will get a RuntimeError.
-    from . import mouse_review
-except RuntimeError:
-    pass
-
 def onEnabledChanged(gsetting, key):
     try:
         enabled = gsetting.get_boolean(key)
@@ -163,6 +158,8 @@ def setLocusOfFocus(event, obj, notifyScript=True, force=False):
         return
 
     if orca_state.activeScript:
+        msg = "ORCA: Active script is: %s" % orca_state.activeScript
+        debug.println(debug.LEVEL_INFO, msg, True)
         if orca_state.activeScript.utilities.isZombie(obj):
             msg = "ERROR: New locusOfFocus (%s) is zombie. Not updating." % obj
             debug.println(debug.LEVEL_INFO, msg, True)
@@ -186,103 +183,6 @@ def setLocusOfFocus(event, obj, notifyScript=True, force=False):
 
     orca_state.activeScript.locusOfFocusChanged(event, oldFocus, orca_state.locusOfFocus)
 
-########################################################################
-#                                                                      #
-# METHODS FOR PRE-PROCESSING AND MASSAGING KEYBOARD EVENTS.            #
-#                                                                      #
-########################################################################
-
-_orcaModifierPressed = False
-
-def _processKeyboardEvent(event):
-    """The primary key event handler for Orca.  Keeps track of various
-    attributes, such as the lastInputEvent.  Also does key echo as well
-    as any local keybindings before passing the event on to the active
-    script.  This method is called synchronously from the AT-SPI registry
-    and should be performant.  In addition, it must return True if it has
-    consumed the event (and False if not).
-
-    Arguments:
-    - event: an AT-SPI DeviceEvent
-
-    Returns True if the event should be consumed.
-    """
-    global _orcaModifierPressed
-
-    keyboardEvent = KeyboardEvent(event)
-    debug.println(debug.LEVEL_FINE, keyboardEvent.toString())
-
-    # Weed out duplicate and otherwise bogus events.
-    # TODO - JD: Be sure these are the right values to return
-    if not keyboardEvent.timestamp:
-        debug.println(debug.LEVEL_FINE, "IGNORING EVENT: NO TIMESTAMP")
-        return False
-    if keyboardEvent == orca_state.lastInputEvent:
-        debug.println(debug.LEVEL_FINE, "IGNORING EVENT: DUPLICATE")
-        return False
-
-    # Figure out what we've got.
-    isOrcaModifier = keyboardEvent.isOrcaModifier()
-    isPressedEvent = keyboardEvent.isPressedKey()
-    if isOrcaModifier:
-        _orcaModifierPressed = isPressedEvent
-    if _orcaModifierPressed:
-        keyboardEvent.modifiers |= keybindings.ORCA_MODIFIER_MASK
-
-    # Update our state.
-    orca_state.lastInputEvent = keyboardEvent
-    if not keyboardEvent.isModifierKey():
-        keyboardEvent.setClickCount()
-        orca_state.lastNonModifierKeyEvent = keyboardEvent
-
-    # Echo it based on what it is and the user's settings.
-    script = orca_state.activeScript
-    if not script:
-        debug.println(debug.LEVEL_FINE, "IGNORING EVENT DUE TO NO SCRIPT")
-        return False
-
-    if isPressedEvent:
-        if not orca_state.activeWindow:
-            orca_state.activeWindow = script.utilities.activeWindow()
-        script.presentationInterrupt()
-    script.presentKeyboardEvent(keyboardEvent)
-    if keyboardEvent.isModifierKey() and not isOrcaModifier:
-        return orca_state.learnModeEnabled
- 
-    # Special modes.
-    if not isPressedEvent and keyboardEvent.event_string == "Escape":
-        script.exitLearnMode(keyboardEvent)
-    if orca_state.learnModeEnabled and not keyboardEvent.modifiers:
-        if keyboardEvent.event_string == "F1":
-            orca_state.learnModeEnabled = False
-            return helpForOrca()
-        if isPressedEvent and keyboardEvent.event_string in ["F2", "F3"]:
-            return script.listOrcaShortcuts(keyboardEvent)
-    if orca_state.capturingKeys:
-        return False
-    if notification_messages.listNotificationMessagesModeEnabled:
-        return notification_messages.listNotificationMessages(script, keyboardEvent)
-
-    # See if the event manager wants it (i.e. it is bound to a command.
-    if _eventManager.processKeyboardEvent(keyboardEvent):
-        return True
-
-    # Do any needed xmodmap crap.
-    global _restoreOrcaKeys
-    if not isPressedEvent:
-        if keyboardEvent.event_string in settings.orcaModifierKeys \
-           and orca_state.bypassNextCommand:
-            _restoreXmodmap()
-            _restoreOrcaKeys = True
-        elif _restoreOrcaKeys and not orca_state.bypassNextCommand:
-            _createOrcaXmodmap()
-            _restoreOrcaKeys = False
-    elif not keyboardEvent.isModifierKey():
-        _orcaModifierPressed = False
-        orca_state.bypassNextCommand = False
- 
-    return isOrcaModifier or orca_state.learnModeEnabled
-
 ########################################################################
 #                                                                      #
 # METHODS FOR PRE-PROCESSING AND MASSAGING BRAILLE EVENTS.             #
@@ -323,6 +223,23 @@ def _processBrailleEvent(event):
 #                                                                      #
 ########################################################################
 
+def updateKeyMap(keyboardEvent):
+    """Unsupported convenience method to call sad hacks which should go away."""
+
+    global _restoreOrcaKeys
+    if keyboardEvent.isPressedKey():
+        return
+
+    if keyboardEvent.event_string in settings.orcaModifierKeys \
+       and orca_state.bypassNextCommand:
+        _restoreXmodmap()
+        _restoreOrcaKeys = True
+        return
+
+    if _restoreOrcaKeys and not orca_state.bypassNextCommand:
+        _createOrcaXmodmap()
+        _restoreOrcaKeys = False
+
 def _setXmodmap(xkbmap):
     """Set the keyboard map using xkbcomp."""
     p = subprocess.Popen(['xkbcomp', '-w0', '-', os.environ['DISPLAY']],
@@ -334,15 +251,14 @@ def _setCapsLockAsOrcaModifier(enable):
     interpretCapsLineProg = re.compile(
         r'^\s*interpret\s+Caps[_+]Lock[_+]AnyOfOrNone\s*\(all\)\s*{\s*$', re.I)
     capsModLineProg = re.compile(
-        r'^\s*action\s*=\s*SetMods\s*\(\s*modifiers\s*=\s*Lock\s*,\s*clearLocks\s*\)\s*;\s*$', re.I)
+        r'^\s*action\s*=\s*NoAction\s*\(\s*\)\s*;\s*$', re.I)
     normalCapsLineProg = re.compile(
         r'^\s*action\s*=\s*LockMods\s*\(\s*modifiers\s*=\s*Lock\s*\)\s*;\s*$', re.I)
     normalCapsLine = '        action= LockMods(modifiers=Lock);'
-    capsModLine =    '        action= SetMods(modifiers=Lock,clearLocks);'
+    capsModLine =    '        action= NoAction();'
     lines = _originalXmodmap.decode('UTF-8').split('\n')
     foundCapsInterpretSection = False
-    for i in range(len(lines)):
-        line = lines[i]
+    for i, line in enumerate(lines):
         if not foundCapsInterpretSection:
             if interpretCapsLineProg.match(line):
                 foundCapsInterpretSection = True
@@ -464,16 +380,14 @@ def loadUserSettings(script=None, inputEvent=None, skipReloadMessage=False):
             msg = 'ORCA: Could not initialize connection to braille.'
             debug.println(debug.LEVEL_WARNING, msg, True)
 
+    if _settingsManager.getSetting('enableMouseReview'):
+        mouse_review.reviewer.activate()
+    else:
+        mouse_review.reviewer.deactivate()
+
     if _settingsManager.getSetting('enableSound'):
         player.init()
 
-    # I'm not sure where else this should go. But it doesn't really look
-    # right here.
-    try:
-        mouse_review.mouse_reviewer.toggle(on=settings.enableMouseReview)
-    except NameError:
-        pass
-
     global _orcaModifiers
     custom = [k for k in settings.orcaModifierKeys if k not in _orcaModifiers]
     _orcaModifiers += custom
@@ -513,7 +427,7 @@ def _showPreferencesUI(script, prefs):
     orca_state.orcaOS.showGUI()
 
 def showAppPreferencesGUI(script=None, inputEvent=None):
-    """Displays the user interace to configure the settings for a
+    """Displays the user interface to configure the settings for a
     specific applications within Orca and set up those app-specific
     user preferences using a GUI.
 
@@ -530,7 +444,7 @@ def showAppPreferencesGUI(script=None, inputEvent=None):
     return True
 
 def showPreferencesGUI(script=None, inputEvent=None):
-    """Displays the user interace to configure Orca and set up
+    """Displays the user interface to configure Orca and set up
     user preferences using a GUI.
 
     Returns True to indicate the input event has been consumed.
@@ -547,6 +461,7 @@ def helpForOrca(script=None, inputEvent=None, page=""):
 
     Returns True to indicate the input event has been consumed.
     """
+    orca_state.learnModeEnabled = False
     uri = "help:orca"
     if page:
         uri += "?%s" % page
@@ -567,7 +482,7 @@ def quitOrca(script=None, inputEvent=None):
     return True
 
 def showFindGUI(script=None, inputEvent=None):
-    """Displays the user interace to perform an Orca Find.
+    """Displays the user interface to perform an Orca Find.
 
     Returns True to indicate the input event has been consumed.
     """
@@ -605,7 +520,6 @@ def init(registry):
         signal.alarm(settings.timeoutTime)
 
     loadUserSettings()
-    _eventManager.registerKeystrokeListener(_processKeyboardEvent)
 
     if settings.timeoutCallback and (settings.timeoutTime > 0):
         signal.alarm(0)
@@ -642,14 +556,9 @@ def start(registry, cacheValues):
     if cacheValues:
         pyatspi.setCacheLevel(pyatspi.CACHE_PROPERTIES)
 
-    try:
-        Gdk.notify_startup_complete()
-    except:
-        msg = 'ERROR: Exception calling Gdk.notification_startup_complete()'
-        debug.println(debug.LEVEL_INFO, msg, True)
-    else:
-        msg = 'ORCA: Startup complete notification made'
-        debug.println(debug.LEVEL_INFO, msg, True)
+    Gdk.notify_startup_complete()
+    msg = 'ORCA: Startup complete notification made'
+    debug.println(debug.LEVEL_INFO, msg, True)
 
     debug.println(debug.LEVEL_INFO, 'ORCA: Starting registry', True)
     registry.start(gil=False)
@@ -696,8 +605,6 @@ def shutdown(script=None, inputEvent=None):
     orca_state.activeScript.presentMessage(messages.STOP_ORCA)
 
     _scriptManager.deactivate()
-
-    _eventManager.deregisterKeystrokeListener(_processKeyboardEvent)
     _eventManager.deactivate()
 
     # Shutdown all the other support.
@@ -763,11 +670,10 @@ def shutdownOnSignal(signum, frame):
     if not cleanExit:
         die(EXIT_CODE_HANG)
 
-def abortOnSignal(signum, frame):
-    debug.println(debug.LEVEL_ALL,
-                  "Aborting due to signal = %d" \
-                  % signum)
-    die(signum)
+def crashOnSignal(signum, frame):
+    signal.signal(signum, signal.SIG_DFL)
+    _restoreXmodmap(_orcaModifiers)
+    os.kill(os.getpid(), signum)
 
 def main(cacheValues=True):
     """The main entry point for Orca.  The exit codes for Orca will
@@ -776,6 +682,11 @@ def main(cacheValues=True):
     an exit code of 0 means normal completion and an exit code of 50
     means Orca exited because of a hang."""
 
+    if debug.debugFile and os.path.exists(debug.debugFile.name):
+        faulthandler.enable(file=debug.debugFile, all_threads=False)
+    else:
+        faulthandler.enable(all_threads=False)
+
     # Method to call when we think something might be hung.
     #
     settings.timeoutCallback = timeout
@@ -786,7 +697,7 @@ def main(cacheValues=True):
     signal.signal(signal.SIGINT, shutdownOnSignal)
     signal.signal(signal.SIGTERM, shutdownOnSignal)
     signal.signal(signal.SIGQUIT, shutdownOnSignal)
-    signal.signal(signal.SIGSEGV, abortOnSignal)
+    signal.signal(signal.SIGSEGV, crashOnSignal)
 
     if not _settingsManager.isAccessibilityEnabled():
         _settingsManager.setAccessibility(True)
@@ -804,7 +715,18 @@ def main(cacheValues=True):
     if script:
         window = script.utilities.activeWindow()
         if window and not orca_state.locusOfFocus:
+            try:
+                app = window.getApplication()
+            except:
+                msg = "ORCA: Exception getting app for %s" % window
+                debug.println(debug.LEVEL_INFO, msg, True)
+            else:
+                script = _scriptManager.getScript(app, window)
+                _scriptManager.setActiveScript(script, "Launching.")
             setLocusOfFocus(None, window)
+            focusedObject = script.utilities.focusedObject(window)
+            if focusedObject:
+                setLocusOfFocus(None, focusedObject)
 
     try:
         start(pyatspi.Registry, cacheValues) # waits until we stop the registry
