@@ -161,6 +161,14 @@ class Script(default.Script):
                 self.inputEventHandlers.get("enableStickyBrowseModeHandler"),
                 3))
 
+        keyBindings.add(
+            keybindings.KeyBinding(
+                "",
+                keybindings.defaultModifierMask,
+                keybindings.NO_MODIFIER_MASK,
+                self.inputEventHandlers.get("toggleLayoutModeHandler")))
+
+
         layout = _settingsManager.getSetting('keyboardLayout')
         if layout == settings.GENERAL_KEYBOARD_LAYOUT_DESKTOP:
             key = "KP_Multiply"
@@ -226,6 +234,11 @@ class Script(default.Script):
                 Script.enableStickyBrowseMode,
                 cmdnames.SET_BROWSE_MODE_STICKY)
 
+        self.inputEventHandlers["toggleLayoutModeHandler"] = \
+            input_event.InputEventHandler(
+                Script.toggleLayoutMode,
+                cmdnames.TOGGLE_LAYOUT_MODE)
+
     def getBookmarks(self):
         """Returns the "bookmarks" class for this script."""
 
@@ -485,6 +498,15 @@ class Script(default.Script):
             if event.source.getRole() == pyatspi.ROLE_LINK:
                 return False
 
+        if event.type.startswith('object:children-changed'):
+            try:
+                role = event.any_data.getRole()
+            except:
+                pass
+            else:
+                if role == pyatspi.ROLE_DIALOG:
+                    return False
+
         return super().skipObjectEvent(event)
 
     def consumesKeyboardEvent(self, keyboardEvent):
@@ -574,6 +596,7 @@ class Script(default.Script):
                     continue
 
                 obj, startOffset, endOffset, text = content
+                eventsynthesizer.scrollIntoView(obj)
                 utterances = self.speechGenerator.generateContents(
                     [content], eliminatePauses=True, priorObj=priorObj)
                 priorObj = obj
@@ -610,11 +633,14 @@ class Script(default.Script):
         if not (text and text.getNSelections()):
             return
 
-        context = self.utilities.getCaretContext(documentFrame=None)
+        document = self.utilities.getDocumentForObject(obj)
+        if not document:
+            return
 
+        context = self.utilities.getCaretContext(documentFrame=document)
         start, end = text.getSelection(0)
         offset = max(offset, start)
-        self.utilities.setCaretContext(obj, offset, documentFrame=None)
+        self.utilities.setCaretContext(obj, offset, documentFrame=document)
         if end - start < _settingsManager.getSetting('findResultsMinimumLength'):
             return
 
@@ -624,7 +650,7 @@ class Script(default.Script):
 
         if self._madeFindAnnouncement \
            and verbosity == settings.FIND_SPEAK_IF_LINE_CHANGED \
-           and not self.utilities.contextsAreOnSameLine(context, (obj, offset)):
+           and self.utilities.contextsAreOnSameLine(context, (obj, offset)):
             return
 
         contents = self.utilities.getLineContentsAtOffset(obj, offset)
@@ -757,6 +783,11 @@ class Script(default.Script):
             debug.println(debug.LEVEL_INFO, msg, True)
             return True
 
+        if self._inFocusMode and obj and obj.getRole() == pyatspi.ROLE_RADIO_BUTTON:
+            msg = "WEB: Staying in focus mode due to role of %s" % obj
+            debug.println(debug.LEVEL_INFO, msg, True)
+            return True
+
         msg = "WEB: Not using focus mode for %s due to lack of cause" % obj
         debug.println(debug.LEVEL_INFO, msg, True)
         return False
@@ -810,13 +841,13 @@ class Script(default.Script):
     def sayLine(self, obj):
         """Speaks the line at the current caret position."""
 
-        if not (self._lastCommandWasCaretNav or self._lastCommandWasStructNav) \
-           and not self.utilities.isContentEditableWithEmbeddedObjects(obj):
+        isEditable = self.utilities.isContentEditableWithEmbeddedObjects(obj)
+        if not (self._lastCommandWasCaretNav or self._lastCommandWasStructNav) and not isEditable:
             super().sayLine(obj)
             return
 
         priorObj = None
-        if self._lastCommandWasCaretNav:
+        if self._lastCommandWasCaretNav or isEditable:
             priorObj, priorOffset = self.utilities.getPriorContext()
 
         obj, offset = self.utilities.getCaretContext(documentFrame=None)
@@ -824,15 +855,35 @@ class Script(default.Script):
         self.speakContents(contents, priorObj=priorObj)
 
     def presentObject(self, obj, **args):
-        priorObj = None
+        if not self.utilities.inDocumentContent(obj):
+            super().presentObject(obj, **args)
+            return
+
+        priorObj = args.get("priorObj")
         if self._lastCommandWasCaretNav or args.get("includeContext"):
             priorObj, priorOffset = self.utilities.getPriorContext()
 
+        if obj.getRole() == pyatspi.ROLE_ENTRY:
+            utterances = self.speechGenerator.generateSpeech(obj, priorObj=priorObj)
+            speech.speak(utterances)
+            self.updateBraille(obj)
+            return
+
         offset = args.get("offset", 0)
         contents = self.utilities.getObjectContentsAtOffset(obj, offset)
         self.displayContents(contents)
-        self.speakContents(contents, priorObj=priorObj)
+        self.speakContents(contents, **args)
  
+    def updateBrailleForNewCaretPosition(self, obj):
+        """Try to reposition the cursor without having to do a full update."""
+
+        text = self.utilities.queryNonEmptyText(obj)
+        if text and self.EMBEDDED_OBJECT_CHARACTER in text.getText(0, -1):
+            self.updateBraille(obj)
+            return
+
+        super().updateBrailleForNewCaretPosition(obj)
+
     def updateBraille(self, obj, **args):
         """Updates the braille display to show the given object."""
 
@@ -847,10 +898,12 @@ class Script(default.Script):
             super().updateBraille(obj, **args)
             return
 
+        isContentEditable = self.utilities.isContentEditableWithEmbeddedObjects(obj)
+
         if not self._lastCommandWasCaretNav \
            and not self._lastCommandWasStructNav \
+           and not isContentEditable \
            and not self.utilities.isPlainText() \
-           and not self.utilities.isContentEditableWithEmbeddedObjects(obj) \
            and not self.utilities.lastInputEventWasCaretNavWithSelection():
             msg = "WEB: updating braille for unhandled navigation type %s" % obj
             debug.println(debug.LEVEL_INFO, msg, True)
@@ -858,6 +911,11 @@ class Script(default.Script):
             return
 
         obj, offset = self.utilities.getCaretContext(documentFrame=None)
+        if offset > 0 and isContentEditable:
+            text = self.utilities.queryNonEmptyText(obj)
+            if text:
+                offset = min(offset, text.characterCount)
+
         contents = self.utilities.getLineContentsAtOffset(obj, offset)
         self.displayContents(contents)
 
@@ -1052,6 +1110,14 @@ class Script(default.Script):
         self._focusModeIsSticky = True
         self._browseModeIsSticky = False
 
+    def toggleLayoutMode(self, inputEvent):
+        layoutMode = not _settingsManager.getSetting('layoutMode')
+        if layoutMode:
+            self.presentMessage(messages.MODE_LAYOUT)
+        else:
+            self.presentMessage(messages.MODE_OBJECT)
+        _settingsManager.setSetting('layoutMode', layoutMode)
+
     def togglePresentationMode(self, inputEvent):
         [obj, characterOffset] = self.utilities.getCaretContext()
         if self._inFocusMode:
@@ -1092,30 +1158,31 @@ class Script(default.Script):
             debug.println(debug.LEVEL_INFO, msg, True)
             return False
 
-        if oldFocus and self.utilities.isZombie(oldFocus):
-            msg = "WEB: Old focus is Zombie: %s. Clearing oldFocus." % oldFocus
-            debug.println(debug.LEVEL_INFO, msg, True)
-            oldFocus = None
-
         caretOffset = 0
-        if not oldFocus or self.utilities.inFindToolbar(oldFocus):
+        if not oldFocus or self.utilities.inFindToolbar(oldFocus) \
+           or (self.utilities.isDocument(newFocus) and oldFocus == orca_state.activeWindow):
             contextObj, contextOffset = self.utilities.getCaretContext()
             if contextObj and not self.utilities.isZombie(contextObj):
                 newFocus, caretOffset = contextObj, contextOffset
 
-        if newFocus.getRole() == pyatspi.ROLE_UNKNOWN:
+        if newFocus.getRole() in [pyatspi.ROLE_UNKNOWN, pyatspi.ROLE_REDUNDANT_OBJECT]:
             msg = "WEB: Event source has bogus role. Likely browser bug."
             debug.println(debug.LEVEL_INFO, msg, True)
             newFocus, offset = self.utilities.findFirstCaretContext(newFocus, 0)
 
         text = self.utilities.queryNonEmptyText(newFocus)
-        if text and (0 <= text.caretOffset < text.characterCount):
+        if text and (0 <= text.caretOffset <= text.characterCount):
             caretOffset = text.caretOffset
 
         self.utilities.setCaretContext(newFocus, caretOffset)
         self.updateBraille(newFocus)
 
-        if self.utilities.isAnchor(newFocus):
+        if self.utilities.isContentEditableWithEmbeddedObjects(newFocus):
+            msg = "WEB: New focus %s content editable. Generating line contents." % newFocus
+            debug.println(debug.LEVEL_INFO, msg, True)
+            contents = self.utilities.getLineContentsAtOffset(newFocus, caretOffset)
+            utterances = self.speechGenerator.generateContents(contents)
+        elif self.utilities.isAnchor(newFocus):
             msg = "WEB: New focus %s is anchor. Generating line contents." % newFocus
             debug.println(debug.LEVEL_INFO, msg, True)
             contents = self.utilities.getLineContentsAtOffset(newFocus, 0)
@@ -1125,6 +1192,11 @@ class Script(default.Script):
             debug.println(debug.LEVEL_INFO, msg, True)
             contents = self.utilities.getLineContentsAtOffset(newFocus, caretOffset)
             utterances = self.speechGenerator.generateContents(contents)
+        elif self.utilities.isFocusedWithMathChild(newFocus):
+            msg = "WEB: New focus %s has math child. Generating line contents." % newFocus
+            debug.println(debug.LEVEL_INFO, msg, True)
+            contents = self.utilities.getLineContentsAtOffset(newFocus, caretOffset)
+            utterances = self.speechGenerator.generateContents(contents)
         elif newFocus.getRole() == pyatspi.ROLE_HEADING:
             msg = "WEB: New focus %s is heading. Generating object contents." % newFocus
             debug.println(debug.LEVEL_INFO, msg, True)
@@ -1202,8 +1274,6 @@ class Script(default.Script):
             debug.println(debug.LEVEL_INFO, msg, True)
             return True
 
-        self._loadingDocumentContent = event.detail1
-
         obj, offset = self.utilities.getCaretContext()
         if not obj or self.utilities.isZombie(obj):
             self.utilities.clearCaretContext()
@@ -1218,12 +1288,20 @@ class Script(default.Script):
             else:
                 self.presentMessage(messages.PAGE_LOADING_END)
 
+        activeDocument = self.utilities.activeDocument()
+        if activeDocument and activeDocument != event.source:
+            msg = "WEB: Ignoring: Event source is not active document"
+            debug.println(debug.LEVEL_INFO, msg, True)
+            return True
+
+        self._loadingDocumentContent = event.detail1
         if event.detail1:
             return True
 
         self.utilities.clearCachedObjects()
 
         if _settingsManager.getSetting('pageSummaryOnLoad') and shouldPresent:
+            obj = obj or event.source
             msg = "WEB: Getting page summary for obj %s" % obj
             debug.println(debug.LEVEL_INFO, msg, True)
             summary = self.utilities.getPageSummary(obj)
@@ -1334,18 +1412,22 @@ class Script(default.Script):
             msg = "WEB: Event handled: Last command was mouse button"
             debug.println(debug.LEVEL_INFO, msg, True)
             self.utilities.setCaretContext(event.source, event.detail1)
-            orca.setLocusOfFocus(event, event.source)
+            notify = not self.utilities.isEntryDescendant(event.source)
+            orca.setLocusOfFocus(event, event.source, notify)
+            if orca_state.locusOfFocus == event.source:
+                self.updateBraille(event.source)
             return True
 
         if self.utilities.inFindToolbar():
-            if not self._madeFindAnnouncement:
-                msg = "WEB: Event handled: Presenting find results"
-                debug.println(debug.LEVEL_INFO, msg, True)
-                self.presentFindResults(event.source, event.detail1)
-            else:
-                self.utilities.setCaretContext(event.source, event.detail1)
-                msg = "WEB: Event handled: Setting context to source and offset"
-                debug.println(debug.LEVEL_INFO, msg, True)
+            msg = "WEB: Event handled: Presenting find results"
+            debug.println(debug.LEVEL_INFO, msg, True)
+            self.presentFindResults(event.source, event.detail1)
+            self._saveFocusedObjectInfo(orca_state.locusOfFocus)
+            return True
+
+        if self.utilities.inContextMenu():
+            msg = "WEB: Event ignored: In context menu"
+            debug.println(debug.LEVEL_INFO, msg, True)
             return True
 
         if self.utilities.eventIsAutocompleteNoise(event):
@@ -1378,14 +1460,21 @@ class Script(default.Script):
             msg = "WEB: Event handled: Caret moved due to scrolling"
             debug.println(debug.LEVEL_INFO, msg, True)
             self.utilities.setCaretContext(obj, offset)
-            orca.setLocusOfFocus(event, obj, force=self.utilities.isPlainText())
+            orca.setLocusOfFocus(event, obj, force=True)
             return True
 
         if self.utilities.isContentEditableWithEmbeddedObjects(event.source):
             msg = "WEB: In content editable with embedded objects"
             debug.println(debug.LEVEL_INFO, msg, True)
+            if not self.utilities.eventIsFromLocusOfFocusDocument(event):
+                msg = "WEB: Event ignored: Not from locus of focus document"
+                debug.println(debug.LEVEL_INFO, msg, True)
+                return True
+
             self.utilities.setCaretContext(obj, offset)
-            orca.setLocusOfFocus(event, event.source, True)
+            notify = not self.utilities.lastInputEventWasCharNav() \
+                     and not self.utilities.isEntryDescendant(obj)
+            orca.setLocusOfFocus(event, event.source, notify)
             return False
 
         text = self.utilities.queryNonEmptyText(event.source)
@@ -1651,9 +1740,10 @@ class Script(default.Script):
 
             obj, offset = self.utilities.searchForCaretContext(event.source)
             if obj:
+                notify = self.utilities.inFindToolbar(orca_state.locusOfFocus)
                 msg = "WEB: Updating focus and context to %s, %i" % (obj, offset)
                 debug.println(debug.LEVEL_INFO, msg, True)
-                orca.setLocusOfFocus(event, obj, False)
+                orca.setLocusOfFocus(event, obj, notify)
                 self.utilities.setCaretContext(obj, offset)
             else:
                 msg = "WEB: Search for caret context failed"
@@ -1748,8 +1838,10 @@ class Script(default.Script):
             return True
 
         if self.utilities.eventIsChromePageSwitchNoise(event):
-            msg = "WEB: Ignoring event believed to be chrome page switch noise"
+            msg = "WEB: Event believed to be chrome page switch"
             debug.println(debug.LEVEL_INFO, msg, True)
+            if event.detail1:
+                self.presentObject(event.source)
             return True
 
         if not self.utilities.inDocumentContent(event.source):
@@ -1788,6 +1880,11 @@ class Script(default.Script):
             debug.println(debug.LEVEL_INFO, msg, True)
             return True
 
+        if not self.utilities.eventIsFromLocusOfFocusDocument(event):
+            msg = "WEB: Event ignored: Not from locus of focus document"
+            debug.println(debug.LEVEL_INFO, msg, True)
+            return True
+
         obj, offset = self.utilities.getCaretContext()
         ancestor = self.utilities.commonAncestor(obj, event.source)
         if ancestor and self.utilities.isTextBlockElement(ancestor):
@@ -1842,6 +1939,11 @@ class Script(default.Script):
 
         obj, offset = self.utilities.getCaretContext(getZombieReplicant=False)
         if self.utilities.isZombie(obj):
+            if self.utilities.isLink(obj):
+                msg = "WEB: Focused link deleted. Taking no further action."
+                debug.println(debug.LEVEL_INFO, msg, True)
+                return True
+
             obj, offset = self.utilities.getCaretContext(getZombieReplicant=True)
             if obj:
                 orca.setLocusOfFocus(event, obj, notifyScript=False)
@@ -1860,8 +1962,8 @@ class Script(default.Script):
             debug.println(debug.LEVEL_INFO, msg, True)
             self.structuralNavigation.clearCache(document)
 
-        state = event.source.getState()
-        if not state.contains(pyatspi.STATE_EDITABLE):
+        if not event.source.getState().contains(pyatspi.STATE_EDITABLE) \
+           and not self.utilities.isContentEditableWithEmbeddedObjects(event.source):
             if self._inMouseOverObject \
                and self.utilities.isZombie(self._lastMouseOverObject):
                 msg = "WEB: Restoring pre-mouseover context"
@@ -1959,13 +2061,6 @@ class Script(default.Script):
             debug.println(debug.LEVEL_INFO, msg, True)
             return False
 
-        if self.utilities.inFindToolbar():
-            msg = "WEB: Event handled: Presenting find results"
-            debug.println(debug.LEVEL_INFO, msg, True)
-            self.presentFindResults(event.source, -1)
-            self._saveFocusedObjectInfo(orca_state.locusOfFocus)
-            return True
-
         if not self.utilities.inDocumentContent(orca_state.locusOfFocus):
             msg = "WEB: Event ignored: locusOfFocus (%s) is not in document content" \
                   % orca_state.locusOfFocus
