@@ -47,8 +47,11 @@ from . import speechserver
 from . import settings
 from . import orca_state
 from . import punctuation_settings
+from . import settings_manager
 from .acss import ACSS
 
+_settingsManager = settings_manager.getManager()
+
 try:
     import speechd
 except:
@@ -62,8 +65,8 @@ else:
     else:
         _speechd_version_ok = True
 
-PUNCTUATION = re.compile('[^\w\s]', re.UNICODE)
-ELLIPSIS = re.compile('(\342\200\246|(?<!\.)\.{3,4}(?=(\s|\Z)))')
+PUNCTUATION = re.compile(r'[^\w\s]', re.UNICODE)
+ELLIPSIS = re.compile('(\342\200\246|(?<!\\.)\\.{3,4}(?=(\\s|\\Z)))')
 
 class SpeechServer(speechserver.SpeechServer):
     # See the parent class for documentation.
@@ -97,7 +100,7 @@ class SpeechServer(speechserver.SpeechServer):
         """
         if serverId not in cls._active_servers:
             cls(serverId)
-        # Don't return the instance, unless it is succesfully added
+        # Don't return the instance, unless it is successfully added
         # to `_active_Servers'.
         return cls._active_servers.get(serverId)
 
@@ -145,7 +148,7 @@ class SpeechServer(speechserver.SpeechServer):
             speechd.CallbackType.BEGIN: speechserver.SayAllContext.PROGRESS,
             speechd.CallbackType.CANCEL: speechserver.SayAllContext.INTERRUPTED,
             speechd.CallbackType.END: speechserver.SayAllContext.COMPLETED,
-           #speechd.CallbackType.INDEX_MARK:speechserver.SayAllContext.PROGRESS,
+            speechd.CallbackType.INDEX_MARK:speechserver.SayAllContext.PROGRESS,
             }
 
         self._default_voice_name = guilabels.SPEECH_DEFAULT_VOICE % serverId
@@ -169,6 +172,7 @@ class SpeechServer(speechserver.SpeechServer):
         self._current_voice_properties = {}
         mode = self._PUNCTUATION_MODE_MAP[settings.verbalizePunctuationStyle]
         client.set_punctuation(mode)
+        client.set_data_mode(speechd.DataMode.SSML)
 
     def updateCapitalizationStyle(self):
         """Updates the capitalization style used by the speech server."""
@@ -309,6 +313,10 @@ class SpeechServer(speechserver.SpeechServer):
         Returns a text string with the punctuation symbols adjusted accordingly.
         """
 
+        style = _settingsManager.getSetting("verbalizePunctuationStyle")
+        if style == settings.PUNCTUATION_STYLE_NONE:
+            return oldText
+
         spokenEllipsis = messages.SPOKEN_ELLIPSIS + " "
         newText = re.sub(ELLIPSIS, spokenEllipsis, oldText)
         symbols = set(re.findall(PUNCTUATION, newText))
@@ -336,6 +344,31 @@ class SpeechServer(speechserver.SpeechServer):
     def _speak(self, text, acss, **kwargs):
         if isinstance(text, ACSS):
             text = ''
+
+        # Mark beginning of words with U+E000 (private use) and record the
+        # string offsets
+        # Note: we need to do this before disturbing the text offsets
+        # Note2: we assume that text mangling below leave U+E000 untouched
+        marks_offsets = []
+        marked_text = ""
+
+        for i in range(len(text)):
+            c = text[i]
+            if c == '\ue000':
+                # Original text already contains U+E000. But syntheses will not
+                # know what to do of it anyway, so discard it
+                continue
+            marked_text += c
+
+            if (c == ' ' or c == '\u00a0') \
+               and i < len(text) - 1 \
+               and text[i + 1] != ' ' and text[i + 1] != '\u00a0':
+                # Word separation, add a mark
+                marks_offsets.append(i + 1)
+                marked_text += '\ue000'
+
+        text = marked_text
+
         text = self.__addVerbalizedPunctuation(text)
         if orca_state.activeScript:
             text = orca_state.activeScript.\
@@ -351,9 +384,38 @@ class SpeechServer(speechserver.SpeechServer):
         #
         text = text.replace('\n.', '\n')
 
+        # Transcribe to SSML, translating U+E000 into marks
+        # Note: we need to do this after all mangling otherwise the ssml markup
+        # would get mangled too
+        ssml = "<speak>"
+        i = 0
+        for c in text:
+            if c == '\ue000':
+                if i >= len(marks_offsets):
+                    # This is really not supposed to happen
+                    msg = "%uth U+E000 does not have corresponding index" % i
+                    debug.println(debug.LEVEL_WARNING, msg, True)
+                else:
+                    ssml += '<mark name="%u"/>' % marks_offsets[i]
+                i += 1
+            # Disable for now, until speech dispatcher properly parses them (version 0.8.9 or later)
+            #elif c == '"':
+            #  ssml += '&quot;'
+            #elif c == "'":
+            #  ssml += '&apos;'
+            elif c == '<':
+              ssml += '&lt;'
+            elif c == '>':
+              ssml += '&gt;'
+            elif c == '&':
+              ssml += '&amp;'
+            else:
+              ssml += c
+        ssml += "</speak>"
+
         self._apply_acss(acss)
-        self._debug_sd_values("Speaking '%s' " % text)
-        self._send_command(self._client.speak, text, **kwargs)
+        self._debug_sd_values("Speaking '%s' " % ssml)
+        self._send_command(self._client.speak, ssml, **kwargs)
 
     def _say_all(self, iterator, orca_callback):
         """Process another sayAll chunk.
@@ -373,7 +435,9 @@ class SpeechServer(speechserver.SpeechServer):
                 t = self._CALLBACK_TYPE_MAP[callbackType]
                 if t == speechserver.SayAllContext.PROGRESS:
                     if index_mark:
-                        context.currentOffset = int(index_mark)
+                        context.currentOffset = context.startOffset + int(index_mark)
+                        msg = "SPEECH DISPATCHER: Got mark %d / %d-%d" % (context.currentOffset, context.startOffset, context.endOffset)
+                        debug.println(debug.LEVEL_INFO, msg, True)
                     else:
                         context.currentOffset = context.startOffset
                 elif t == speechserver.SayAllContext.COMPLETED:
